@@ -29,7 +29,7 @@
 
 #include <sys/types.h>
 #ifdef HAVE_SYS_SOCKET_H
-#include <sys/socket.h>
+#  include <sys/socket.h>
 #endif // HAVE_SYS_SOCKET_H
 
 #include <mutex>
@@ -37,7 +37,7 @@
 #include <vector>
 #include <random>
 #ifndef NOTHREADS
-#include <future>
+#  include <future>
 #endif // NOTHREADS
 
 #include <openssl/ssl.h>
@@ -45,7 +45,7 @@
 #include <ev.h>
 
 #ifdef HAVE_NEVERBLEED
-#include <neverbleed.h>
+#  include <neverbleed.h>
 #endif // HAVE_NEVERBLEED
 
 #include "shrpx_downstream_connection_pool.h"
@@ -63,11 +63,11 @@ struct TicketKeys;
 class MemcachedDispatcher;
 struct UpstreamAddr;
 
-namespace ssl {
+namespace tls {
 
 class CertLookupTree;
 
-} // namespace ssl
+} // namespace tls
 
 struct OCSPUpdateContext {
   // ocsp response buffer
@@ -84,17 +84,18 @@ struct OCSPUpdateContext {
 };
 
 // SerialEvent is an event sent from Worker thread.
-enum SerialEventType {
-  SEV_NONE,
-  SEV_REPLACE_DOWNSTREAM,
+enum class SerialEventType {
+  NONE,
+  REPLACE_DOWNSTREAM,
 };
 
 struct SerialEvent {
   // ctor for event uses DownstreamConfig
-  SerialEvent(int type, const std::shared_ptr<DownstreamConfig> &downstreamconf)
+  SerialEvent(SerialEventType type,
+              const std::shared_ptr<DownstreamConfig> &downstreamconf)
       : type(type), downstreamconf(downstreamconf) {}
 
-  int type;
+  SerialEventType type;
   std::shared_ptr<DownstreamConfig> downstreamconf;
 };
 
@@ -117,6 +118,7 @@ public:
   struct ev_loop *get_loop() const;
   Worker *get_single_worker() const;
   void add_acceptor(std::unique_ptr<AcceptHandler> h);
+  void delete_acceptor();
   void enable_acceptor();
   void disable_acceptor();
   void sleep_acceptor(ev_tstamp t);
@@ -159,11 +161,11 @@ public:
   const std::vector<SSL_CTX *> &get_indexed_ssl_ctx(size_t idx) const;
 
 #ifdef HAVE_NEVERBLEED
-  void set_neverbleed(std::unique_ptr<neverbleed_t> nb);
-  neverbleed_t *get_neverbleed() const;
+  void set_neverbleed(neverbleed_t *nb);
 #endif // HAVE_NEVERBLEED
 
-  // Send SerialEvent SEV_REPLACE_DOWNSTREAM to this object.
+  // Send SerialEvent SerialEventType::REPLACE_DOWNSTREAM to this
+  // object.
   void send_replace_downstream(
       const std::shared_ptr<DownstreamConfig> &downstreamconf);
   // Internal function to send |ev| to this object.
@@ -174,6 +176,8 @@ public:
   void
   worker_replace_downstream(std::shared_ptr<DownstreamConfig> downstreamconf);
 
+  void set_enable_acceptor_on_ocsp_completion(bool f);
+
 private:
   // Stores all SSL_CTX objects.
   std::vector<SSL_CTX *> all_ssl_ctx_;
@@ -198,7 +202,7 @@ private:
   // Worker instance used when single threaded mode (-n1) is used.
   // Otherwise, nullptr and workers_ has instances of Worker instead.
   std::unique_ptr<Worker> single_worker_;
-  std::unique_ptr<ssl::CertLookupTree> cert_tree_;
+  std::unique_ptr<tls::CertLookupTree> cert_tree_;
   std::unique_ptr<MemcachedDispatcher> tls_ticket_key_memcached_dispatcher_;
   // Current TLS session ticket keys.  Note that TLS connection does
   // not refer to this field directly.  They use TicketKeys object in
@@ -207,7 +211,7 @@ private:
   struct ev_loop *loop_;
   std::vector<std::unique_ptr<AcceptHandler>> acceptors_;
 #ifdef HAVE_NEVERBLEED
-  std::unique_ptr<neverbleed_t> nb_;
+  neverbleed_t *nb_;
 #endif // HAVE_NEVERBLEED
   ev_timer disable_acceptor_timer_;
   ev_timer ocsp_timer_;
@@ -220,6 +224,9 @@ private:
   size_t tls_ticket_key_memcached_fail_count_;
   unsigned int worker_round_robin_cnt_;
   bool graceful_shutdown_;
+  // true if acceptors should be enabled after the initial ocsp update
+  // has finished.
+  bool enable_acceptor_on_ocsp_completion_;
 };
 
 } // namespace shrpx
